from abc import abstractmethod
from math import exp
from typing import Tuple

from scp.base_fluid import BaseFluid

# Ideally the type hints would cause the coefficient returns to return this, but it is
#  bulky and I haven't been able to get a new typing NewType to work properly, so for now
#  they just return a plain Tuple.
# Tuple[
#     Tuple[float, float, float, float],
#     Tuple[float, float, float, float],
#     Tuple[float, float, float, float],
#     Tuple[float, float, float],
#     Tuple[float, float],
#     float
# ]


class BaseMelinder(BaseFluid):
    """
    A base class for Melinder fluids that provides convenience methods
    that can be accessed in derived classes.

    Melinder, Å. 2010. Properties of Secondary Working Fluids
    for Indirect Systems. 2nd ed. International Institute of Refrigeration.
    """

    _ij_pairs = (
        (0, 0), (0, 1), (0, 2), (0, 3),
        (1, 0), (1, 1), (1, 2), (1, 3),
        (2, 0), (2, 1), (2, 2), (2, 3),
        (3, 0), (3, 1), (3, 2),
        (4, 0), (4, 1),
        (5, 0)
    )

    @abstractmethod
    def coefficient_freezing(self) -> Tuple:
        """
        Abstract method; derived classes should override to return the
        coefficient matrix for freezing point.
        """
        pass

    @abstractmethod
    def coefficient_viscosity(self) -> Tuple:
        """
        Abstract method; derived classes should override to return the
        coefficient matrix for viscosity.
        """
        pass

    @abstractmethod
    def coefficient_specific_heat(self) -> Tuple:
        """
        Abstract method; derived classes should override to return the
        coefficient matrix for specific heat.
        """
        pass

    @abstractmethod
    def coefficient_conductivity(self) -> Tuple:
        """
        Abstract method; derived classes should override to return the
        coefficient matrix for conductivity.
        """
        pass

    @abstractmethod
    def coefficient_density(self) -> Tuple:
        """
        Abstract method; derived classes should override to return the
        coefficient matrix for density.
        """
        pass

    def __init__(
        self, t_min: float, t_max: float, x: float, x_min: float, x_max: float
    ):

        """
        A constructor for the Melinder fluid base class

        @param t_min: Minimum temperature, in degrees Celsius
        @param t_max: Maximum temperature, in degrees Celsius
        @param x: Concentration fraction, from 0 to 1
        @param x_min: Minimum concentration fraction, from 0 to 1
        @param x_max: Maximum concentration fraction, from 0 to 1
        """

        super().__init__(t_min, t_max, x, x_min, x_max)
        self.x_base = None
        self.t_base = None

    def _f_prop_t_freeze(self, c_arr: Tuple, x: float) -> float:
        """
        General worker function to evaluate fluid properties as
        a function of concentration.

        @param c_arr:
        @param x:

        @return:
        """

        x = self._check_concentration(x)

        xxm = (x * 100) - self.x_base
        yym = self.t_base
        x_xm = [xxm**p for p in range(6)]
        y_ym = [yym**p for p in range(4)]

        f_ret = 0.0

        for i, j in BaseMelinder._ij_pairs:
            f_ret += c_arr[i][j] * x_xm[i] * y_ym[j]

        return f_ret

    def _f_prop(self, c_arr: Tuple, temp: float) -> float:
        """
        General worker function to evaluate fluid properties as
        a function of concentration and temperature.

        @param c_arr:
        @param temp:

        @return:
        """

        temp = self._check_temperature(temp)

        xxm = self.x_pct - self.x_base
        yym = temp - self.t_base
        x_xm = [xxm**p for p in range(6)]
        y_ym = [yym**p for p in range(4)]

        f_ret = 0.0

        for i, j in BaseMelinder._ij_pairs:
            f_ret += c_arr[i][j] * x_xm[i] * y_ym[j]

        return f_ret

    def viscosity(self, temp: float) -> float:
        """
        Calculate the dynamic viscosity of the mixture

        @param temp: Fluid temperature, in degrees Celsius
        @return: Dynamic viscosity, in N/m2-s, or Pa-s
        """

        return exp(self._f_prop(self.coefficient_viscosity(), temp)) / 1000.0

    def specific_heat(self, temp: float) -> float:
        """
        Calculates the specific heat of the mixture

        @param temp: Fluid temperature, in degrees Celsius
        @return: Specific heat, in J/kg-K
        """

        return self._f_prop(self.coefficient_specific_heat(), temp)

    def conductivity(self, temp: float) -> float:
        """
        Calculates the thermal conductivity of the mixture

        @param temp: Fluid temperature, in degrees Celsius
        @return: Thermal conductivity, in W/m-K
        """

        return self._f_prop(self.coefficient_conductivity(), temp)

    def density(self, temp: float) -> float:
        """
        Calculates the density of the mixture

        @param temp: Fluid temperature, in degrees Celsius
        @return: Density, in kg/m3
        """

        return self._f_prop(self.coefficient_density(), temp)

    def freeze_point(self, x: float) -> float:
        """
        Calculate the freezing point temperature of the mixture

        @param x: Concentration fraction
        @return Freezing point temperature, in Celsius
        """

        return self._f_prop_t_freeze(self.coefficient_freezing(), x)



import warnings
from abc import ABC, abstractmethod


class BaseFluid(ABC):
    """
    A fluid base class that provides convenience methods that can be accessed
    in derived classes.
    """

    def __init__(
        self,
        t_min: float,
        t_max: float,
        x: float = None,
        x_min: float = None,
        x_max: float = None,
    ):
        """
        A constructor for a base fluid, that takes a concentration as an argument.
        Derived classes can decide how to handle the concentration argument and
        their own constructor interface as needed to construct and manage that
        specific derived class.

        @param t_min: Minimum temperature, in degrees Celsius
        @param t_max: Maximum temperature, in degrees Celsius
        @param x: Concentration fraction, from 0 to 1
        @param x_min: Minimum concentration fraction, from 0 to 1
        @param x_max: Maximum concentration fraction, from 0 to 1
        """

        self.t_min = None
        self.t_max = None
        self._set_temperature_limits(t_min, t_max)

        if type(x) is not type(None):  # noqa: E721
            self.x_min = None
            self.x_max = None
            self.x = None
            self.x_pct = None
            self._set_concentration_limits(x, x_min, x_max)

    @abstractmethod
    def fluid_name(self) -> str:
        """
        An abstract property that needs to return the fluid name in derived fluid classes

        Derived function must be decorated with @property

        @return: string name of the fluid
        """
        pass

    def _set_concentration_limits(self, x: float, x_min: float, x_max: float):
        """
        An internal worker function that checks the given concentration against limits,
        and sets internal variables.

        @param x: The concentration fraction to check, ranging from 0 to 1
        @param x_min: The minimum concentration fraction to allow, ranging from 0 to 1
        @param x_max: The maximum concentration fraction to allow, ranging from 0 to 1
        @return: Nothing
        """

        if x_min >= x_max:
            msg = f'Developer error: Fluid "{self.fluid_name}", x_min is greater than x_max'
            raise ValueError(msg)

        self.x_min = x_min
        self.x_max = x_max
        self.x = self._check_concentration(x)
        self.x_pct = self.x * 100

    def _check_concentration(self, x: float) -> float:
        """
        An internal worker function that checks the given concentration against limits

        @param x: The concentration to check, in percent
        @return: A validated concentration value, in percent
        """

        if x < self.x_min:
            msg = f'Fluid "{self.fluid_name}", concentration must be greater than {self.x_min:0.2f}.\n'
            msg += f"Resetting concentration to {self.x_min:0.2f}."
            warnings.warn(msg)
            return self.x_min
        elif x > self.x_max:
            msg = f'Fluid "{self.fluid_name}", concentration must be less than {self.x_max:0.2f}.\n'
            msg += f"Resetting concentration to {self.x_max:0.2f}."
            warnings.warn(msg)
            return self.x_max
        else:
            return x

    def _set_temperature_limits(self, t_min, t_max) -> None:
        """
        A worker function to override the default temperature min/max values

        @param t_min: The minimum temperature value to allow, in degrees Celsius
        @param t_max: The maximum temperature value to allow, in degrees Celsius
        @return: Nothing
        """

        if t_min >= t_max:
            msg = f'Fluid "{self.fluid_name}", t_min is greater than t_max'
            ValueError(msg)

        self.t_min = t_min
        self.t_max = t_max

    def _check_temperature(self, temp: float) -> float:
        """
        An internal worker function that checks the given temperature against limits

        @param temp: The temperature to check, in degrees Celsius
        @return: A validated temperature value, in degrees Celsius
        """

        if temp < self.t_min:
            msg = f'Fluid "{self.fluid_name}", temperature must be greater than {self.t_min:0.1f}.\n'
            msg += f"Resetting temperature to {self.t_min:0.1f}."
            warnings.warn(msg)
            return self.t_min
        elif temp > self.t_max:
            msg = f'Fluid "{self.fluid_name}", temperature must be less than {self.t_max:0.1f}.\n'
            msg += f"Resetting temperature to {self.t_max:0.1f}."
            warnings.warn(msg)
            return self.t_max
        else:
            return temp

    @abstractmethod
    def freeze_point(self, x: float) -> float:
        """
        Abstract method; derived classes should override the freezing
        point of that fluid

        @param x: Fluid concentration fraction, ranging from 0 to 1
        @return Returns the freezing point of the fluid, in Celsius
        """
        pass

    @staticmethod
    def freeze_point_units() -> str:
        return "C"

    @abstractmethod
    def viscosity(self, temp: float) -> float:
        """
        Abstract method; derived classes should override to return the dynamic
        viscosity of that fluid.

        @param temp: Fluid temperature, in degrees Celsius
        @return: Returns the dynamic viscosity in [Pa-s]
        """
        pass

    @staticmethod
    def viscosity_units() -> str:
        return "Pa-s"

    def mu(self, temp: float) -> float:
        """
        Convenience function for returning the dynamic viscosity by the common letter 'mu'

        @param temp: Fluid temperature, in degrees Celsius
        @return: Returns the dynamic viscosity -- which one is mu in [Pa-s]
        """
        return self.viscosity(temp)

    @abstractmethod
    def specific_heat(self, temp: float) -> float:
        """
        Abstract method; derived classes should override to return the specific heat
        of that fluid.

        @param temp: Fluid temperature, in degrees Celsius
        @return: Returns the specific heat in [J/kg-K]
        """
        pass

    @staticmethod
    def specific_heat_units() -> str:
        return 'J/kg-K'

    def cp(self, temp: float) -> float:
        """
        Convenience function for returning the specific heat by the common shorthand 'cp'

        @param temp: Fluid temperature, in degrees Celsius
        @return: Returns the specific heat in [J/kg-K]
        """
        return self.specific_heat(temp)

    @abstractmethod
    def density(self, temp: float) -> float:
        """
        Abstract method; derived classes should override to return the density
        of that fluid.

        @param temp: Fluid temperature, in degrees Celsius
        @return: Returns the density in [kg/m3]
        """
        pass

    @staticmethod
    def density_units() -> str:
        return 'kg/m3'

    def rho(self, temp: float) -> float:
        """
        Convenience function for returning the density by the common shorthand 'rho'

        @param temp: Fluid temperature, in degrees Celsius
        @return: Returns the density, in [kg/m3]
        """
        return self.density(temp)

    @abstractmethod
    def conductivity(self, temp: float) -> float:
        """
        Abstract method; derived classes should override to return the thermal
        conductivity of that fluid.

        @param temp: Fluid temperature, in degrees Celsius
        @return: Returns the thermal conductivity in [W/m-K]
        """
        pass

    @staticmethod
    def conductivity_units() -> str:
        return 'W/m-K'

    def k(self, temp: float) -> float:
        """
        Convenience function for returning the thermal conductivity by the common shorthand 'k'

        @param temp: Fluid temperature, in degrees Celsius
        @return: Returns the thermal conductivity, in [W/m-K]
        """
        return self.conductivity(temp)

    def prandtl(self, temp: float) -> float:
        """
        Returns the Prandtl number for this fluid

        @param temp: Fluid temperature, in degrees Celsius
        @return: Returns the dimensionless Prandtl number
        """
        return self.cp(temp) * self.mu(temp) / self.k(temp)

    @staticmethod
    def prandtl_units() -> str:
        return '-'

    def pr(self, temp: float = 0.0) -> float:
        """
        Convenience function for returning the Prandtl number by the common shorthand 'pr'

        @param temp: Fluid temperature, in degrees Celsius
        @return: Returns the dimensionless Prandtl number
        """
        return self.prandtl(temp)

    def thermal_diffusivity(self, temp: float) -> float:
        """
        Returns the thermal diffusivity for this fluid

        @param temp: Fluid temperature, in degrees Celsius
        @return: Returns the thermal diffusivity in [m2/s]
        """
        return self.k(temp) / (self.rho(temp) * self.cp(temp))

    @staticmethod
    def thermal_diffusivity_units() -> str:
        return 'm2/s'

    def alpha(self, temp: float) -> float:
        """
        Convenience function for returning the thermal diffusivity by the common shorthand 'alpha'

        @param temp: Fluid temperature, in degrees Celsius
        @return: Returns the thermal diffusivity in [m2/s]
        """
        return self.thermal_diffusivity(temp)



import click
from sys import stderr
from scp import water, ethyl_alcohol, ethylene_glycol, propylene_glycol, methyl_alcohol

# Add this to .bashrc for nice completion:  eval "$(_SCPROP_COMPLETE=bash_source scprop)"

fluids = {
    "water": water.Water,
    "ethyl_alcohol": ethyl_alcohol.EthylAlcohol,
    "ethylene_glycol": ethylene_glycol.EthyleneGlycol,
    "methyl_alcohol": methyl_alcohol.MethylAlcohol,
    "propylene_glycol": propylene_glycol.PropyleneGlycol,
}
fluid_options = click.Choice(list(fluids.keys()))
properties = (
    "viscosity",
    "specific_heat",
    "density",
    "conductivity",
    "prandtl",
    "thermal_diffusivity",
    "freeze_point",
)
prop_options = click.Choice(properties)
x_range = click.FloatRange(min=0.0, max=1.0)


@click.command(name="SecondaryCoolantPropsCommandLine")
@click.option(
    "-f",
    "--fluid",
    type=fluid_options,
    required=True,
    default="water",
    help="Which fluid to use?",
)
@click.option(
    "-x",
    "--concentration",
    type=x_range,
    required=False,
    default=0.0,
    help="Mixture concentration fraction. Default 0.0.",
)
@click.option(
    "-p",
    "--property",
    "fluid_prop",
    type=prop_options,
    required=True,
    help="Which fluid property to evaluate.",
)
@click.option(
    "-t",
    "--temperature",
    type=float,
    required=True,
    help="Fluid temperature, in degrees Celsius.",
)
@click.option(
    "-q",
    "--quick",
    is_flag=True,
    show_default=True,
    default=False,
    help="Just report the value, good for scripts",
)
def cli(fluid: str, concentration: float, fluid_prop: str, temperature: float, quick: bool):
    if concentration == 0.0 and fluid != "water":
        print(
            "Mixture requested, but concentration zero, assuming water and continuing",
            file=stderr,
        )
    elif concentration > 0.0 and fluid == "water":
        print(
            "Pure water requested, but nonzero concentration entered, assuming water and continuing",
            file=stderr,
        )
    if fluid == "water" or concentration == 0.0:
        f = water.Water()
        fluid = "water"
    else:
        f = fluids[fluid](concentration)

    if fluid_prop == "freeze_point":
        value = getattr(f, fluid_prop)(concentration)
    else:
        value = getattr(f, fluid_prop)(temperature)
    units = getattr(f, f"{fluid_prop}_units")()
    if quick:
        print(value)
    else:
        print(f"Fluid:    {fluid}\nProperty: {fluid_prop}\nValue:    {value}\nUnits:    [{units}]")



from scp.base_fluid import BaseFluid


class Water(BaseFluid):
    def __init__(self) -> None:
        """
        This class represents water as a fluid.  The constructor
        does not require any arguments for the pure water fluid.
        """
        super().__init__(0.0, 100.0)

    @property
    def fluid_name(self) -> str:
        """
        Returns the fluid name for this derived fluid.
        @return: "Water"
        """
        return "Water"

    def viscosity(self, temp: float) -> float:
        """
        Returns the viscosity of water.

        @param temp: Fluid temperature, in degrees Celsius
        @return: The dynamic viscosity of water in [Pa-s]
        """

        temp = self._check_temperature(temp)

        am0 = -3.30233
        am1 = 1301
        am2 = 998.333
        am3 = 8.1855
        am4 = 0.00585
        am5 = 1.002
        am6 = -1.3272
        am7 = -0.001053
        am8 = 105
        am10 = 0.68714
        am11 = -0.0059231
        am12 = 2.1249e-05
        am13 = -2.69575e-08

        if temp < 20:
            exponent = am0 + am1 / (am2 + (temp - 20) * (am3 + am4 * (temp - 20)))
            return (10**exponent) * 0.1
        if temp > 100:
            return (am10 + temp * am11 + (temp**2) * am12 + (temp**3) * am13) * 0.001
        return (am5 * 10 ** ((temp - 20) * (am6 + (temp - 20) * am7) / (temp + am8))) * 0.001

    def specific_heat(self, temp: float) -> float:
        """
        Returns the fluid specific heat.

        Specific heat of water at 1 atmosphere, 0 to 100 C.  Equation from linear least-squares
        regression of data from CRC Handbook (op.cit.) page D-174

        @param temp: Fluid temperature, in degrees Celsius
        @return: Specific heat, in [J/kg-K]
        """

        temp = self._check_temperature(temp)

        acp0 = 4.21534
        acp1 = -0.00287819
        acp2 = 7.4729e-05
        acp3 = -7.79624e-07
        acp4 = 3.220424e-09

        return (
            acp0
            + temp * acp1
            + (temp**2) * acp2
            + (temp**3) * acp3
            + (temp**4) * acp4
        ) * 1000

    def freeze_point(self, _=None) -> float:
        """
        Returns the freezing point temperature of water

        @param _: Unused variable
        @return Freezing point temperature, in Celsius
        """
        return 0.0

    def conductivity(self, temp: float) -> float:
        """
        Returns the fluid thermal conductivity for this derived fluid.

        Thermal conductivity equation from linear least-squares fit to data in CRC Handbook (op.cit.), page E-11

        @param temp: Fluid temperature, in degrees Celsius
        @return: Thermal conductivity, in [W/m-K]
        """

        temp = self._check_temperature(temp)

        ak0 = 0.560101
        ak1 = 0.00211703
        ak2 = -1.05172e-05
        ak3 = 1.497323e-08
        ak4 = -1.48553e-11

        return (
            ak0 + temp * ak1 + (temp**2) * ak2 + (temp**3) * ak3 + (temp**4) * ak4
        )

    def density(self, temp: float) -> float:
        """
        Returns the fluid density for this derived fluid.

        Density eq. for water at 1 atm., from CRC Handbook of Chem. & Phys., 61st Edition (1980-1981), p. F-6.

        @param temp: Fluid temperature, in degrees Celsius
        @return: Density, in [kg/m3]
        """

        temp = self._check_temperature(temp)

        ar0 = 999.83952
        ar1 = 16.945176
        ar2 = -0.0079870401
        ar3 = -4.6170461e-05
        ar4 = 1.0556302e-07
        ar5 = -2.8054253e-10
        ar6 = 0.01687985

        return (
            ar0
            + temp * ar1
            + (temp**2) * ar2
            + (temp**3) * ar3
            + (temp**4) * ar4
            + (temp**5) * ar5
        ) / (1 + ar6 * temp)
